package de.hftstuttgart.dtabackend.rest.v1.task;

import de.hftstuttgart.dtabackend.utils.*;
import de.hftstuttgart.dtabackend.models.ResultSummary;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.tika.Tika;
import org.springframework.core.env.Environment;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.multipart.MultipartFile;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import jakarta.servlet.annotation.MultipartConfig;

/**
 * Rest controller for everything related to the TASK files
 */
@RestController
@RequestMapping("/v1/task/*")
@MultipartConfig
public class TaskUpload {
    private static final Logger LOG = LogManager.getLogger(TaskUpload.class);

    private final RepoUtil repoUtil;
    private final Path testTmpPath;
    private final ExecuteTestUtil executeTestUtil;

    public TaskUpload(
        Environment env,
        RepoUtil repoUtil,
        ExecuteTestUtil executeTestUtil
    ) {
        this.repoUtil = repoUtil;
        this.executeTestUtil = executeTestUtil;

        // set path of temporary directory on host and inside our container
        this.testTmpPath = Paths.get(env.getProperty("tests.tmp.dir"));
    }

    @RequestMapping(method = RequestMethod.POST)
    public ResultSummary uploadAndTestFile(@RequestParam("taskFile") MultipartFile taskFileRef,
                                           @RequestParam("assignmentId") String assignmentId
    ) throws IOException, InterruptedException {
        LOG.info("submission for testing received");

        LOG.debug("creating new temporary directory");
        Path workDirectory = Files.createTempDirectory(testTmpPath, "dta");
        LOG.debug(String.format("working dir for test is: %s", workDirectory.toAbsolutePath().toString()));

        // define paths for the test, the submission and where the result is to be expected afterwards
        Path srcPath    = Paths.get(workDirectory.toAbsolutePath().toString(), "src");

        String mimeInfo = new Tika().detect(taskFileRef.getInputStream());
        switch (mimeInfo) {
            case "text/plain":
                LOG.debug("textfile uploaded, searching for dta config");
                // find URI in config file
                String subDir="";
                Matcher config = RegexUtil.extractConfig(taskFileRef.getInputStream(), Pattern.compile(RegexUtil.DTA_SUBMISSIONCONFIGREGEX));
                if(config==null) {
                	config = RegexUtil.extractConfig(taskFileRef.getInputStream(), Pattern.compile(RegexUtil.SUBMISSIONCONFIGREGEX));
                	if(config==null)
                	{
                		throw new RuntimeException("couldn't find repo config for student submission clone");
                	}
                }
                else {
                	subDir=config.group(4);
                }
                LOG.debug("calling repo clone");
                repoUtil.cloneRepository(config, srcPath.toAbsolutePath().toString(), subDir);
                break;

            case "application/zip":
                LOG.debug("zip archive uploaded, extracting content as student submission");
                ArchiveUtil.extractProjectFromZip(taskFileRef.getInputStream(), srcPath.toAbsolutePath());
                break;

            default:
                String msg = String.format("couldn't process uploaded file with mime type %s", mimeInfo);
                LOG.error(msg);
                throw new RuntimeException(msg);
        }

        // run test
        LOG.debug("calling test execution");
        ResultSummary resultSummary = executeTestUtil.runTests(assignmentId, workDirectory);

        if (mimeInfo.equals("text/plain")) {
            LOG.info("check for provided Ticketsystem information");
            UnifiedTicketingUtil.reportResults(taskFileRef.getInputStream(), resultSummary);
        }
        
        taskFileRef.getInputStream().close();

        LOG.info("submission tested successfully");
        return resultSummary;
    }
}
